import { Docker } from "node-docker-api";
import { dockerCaddyName, dockerNetwork, caddyConfPath } from "./const.js";
import {
  CaddyState,
  ContainerData,
  ContainerNetwork,
  ContainerOverride,
  ImageOverride,
  NetworkData,
  NetworkOverride,
} from "./types.js";

export let networks: Record<string, NetworkData> = {};
export let containers: Record<string, ContainerData> = {};

export async function update() {
  const dockerSock = new Docker({ socketPath: "/var/run/docker.sock" });
  await Promise.all([
    updateNetworkData(dockerSock),
    updateContainerData(dockerSock),
  ]);
  const caddyState = await getCaddyState(dockerSock);
  if (caddyState !== "running") await startCaddy(dockerSock, caddyState);
}

async function updateNetworkData(socket: Docker): Promise<void> {
  networks = {};

  for (const network of (await socket.network.list({
    all: true,
  })) as unknown as NetworkOverride[]) {
    const data = network.data;
    // Note: There is no guarantee that Network names don't overlap.
    // There are lots of potential issues connected to non unique names so we will throw an error if this is not the case
    // See the 'CheckDuplicate' argument here:
    // https://docs.docker.com/engine/api/v1.41/#tag/Network/operation/NetworkCreate
    if (data.Name in networks)
      throw new Error(`ERROR: Duplicate docker network name: ${data.Name}`);
    networks[data.Name] = {
      Id: data.Id,
      Name: data.Name,
      Driver: data.Driver,
    };
  }
  return;
}

async function updateContainerData(socket: Docker): Promise<void> {
  containers = {};
  // The Docker-Node Library is somewhat old and has inacurate/missing typing
  for (const container of (await socket.container.list({
    all: true,
  })) as unknown as ContainerOverride[]) {
    const data = container.data;

    const networks = Object.values(data.NetworkSettings.Networks).map(
      (netData: ContainerNetwork) => {
        return [
          netData.NetworkID,
          {
            NetworkID: netData.NetworkID,
            IPAddress: netData.IPAddress,
          },
        ] as [string, ContainerNetwork];
      }
    );
    containers[data.Id] = {
      Id: data.Id,
      Ports: Object.values(data.Ports).map((p) => `${p.PrivatePort}/${p.Type}`),
      Name: data.Names[0] ?? "Unnamed",
      HostNames: data.Names,
      State: data.State,
      Networks: Object.fromEntries(networks),
      Labels: data.Labels,
    };
  }
  return;
}

async function getCaddyState(socket: Docker): Promise<CaddyState> {
  const container = getContainerByName(`/${dockerCaddyName}`);
  if (container) {
    // Check if container is part of the configured network
    const networkCheck = networks[dockerNetwork].Id in container.Networks;
    return networkCheck ? container.State : "wrongnetwork";
  }
  const hasCaddy = (
    (await socket.image.list()) as unknown as ImageOverride[]
  ).some((image) => image?.data?.RepoTags?.includes("caddy:latest"));
  return hasCaddy ? "nocontainer" : "noimage";
}

function getContainerByName(containerName: string) {
  for (const container of Object.values(containers)) {
    if (container.Name === containerName) return container;
  }
  return undefined;
}

function getNetworkByName(networkName: string) {
  for (const network of Object.values(networks)) {
    if (network.Name === networkName) return network;
  }
  return undefined;
}

async function ensureNetwork(socket: Docker) {
  if (!Object.values(networks).some((net) => net.Name === dockerNetwork)) {
    await socket.network.create({ Name: dockerNetwork });
    await updateNetworkData(socket);
  }
}

async function addContainerToNetwork(
  socket: Docker,
  container: ContainerData,
  network: NetworkData
) {
  if (!(network.Id in container.Networks)) {
    await socket.network.get(network.Id).connect({ Container: container.Id });
    await updateContainerData(socket);
  }
}

async function startCaddy(socket: Docker, caddyState: CaddyState) {
  await ensureNetwork(socket);

  switch (caddyState) {
    case "noimage":
      await socket.image.create({}, { fromImage: "caddy", tag: "latest" });
    // Create Container in shared network
    case "nocontainer":
      // Docker Documentation can be found here:
      // https://docs.docker.com/engine/api/v1.41/#tag/Container/operation/ContainerCreate
      await socket.container.create({
        name: dockerCaddyName,
        network: dockerNetwork,
        Image: "caddy",
        HostConfig: {
          PortBindings: {
            "80": [{ HostPort: "80" }],
            "443/tcp": [{ HostIp: "0.0.0.0", HostPort: "443" }],
            "443/udp": [{ HostIp: "0.0.0.0", HostPort: "443" }],
            "2019/tcp": [{ HostIp: "0.0.0.0", HostPort: "2019" }],
          },
          Binds: [`${caddyConfPath}:/etc/caddy/Caddyfile`],
        },
      });
      await updateContainerData(socket);
    case "wrongnetwork":
      await addContainerToNetwork(
        socket,
        getContainerByName(`/${dockerCaddyName}`)!,
        getNetworkByName(dockerNetwork)!
      );
    case "exited":
      const id = getContainerByName(`/${dockerCaddyName}`)?.Id!;
      await socket.container.get(id)?.start();
  }
}

export async function stopCaddy() {
  // Create fresh docker connection (mitigate possible exceptions with the original connection and can be called separately)
  const dockerSock = new Docker({ socketPath: "/var/run/docker.sock" });
  const container = getContainerByName(`/${dockerCaddyName}`);
  if (container?.State === "running")
    await dockerSock.container.get(container.Id).stop();
}
