import {
  accessSync,
  constants,
  copyFileSync,
  existsSync,
  readFileSync,
  writeFileSync,
} from "node:fs";
import { hostConfPath, hostConfBakExt } from "./const.js";
import { HostData, ContainerData } from "./types.js";

let hasChanged = false;
const hostData: HostData = {};
let initialHostData: HostData = {};

function checkFile(path: string) {
  if (!existsSync(path))
    throw new Error(`No such file or directory: ${hostConfPath}`);
  // Throws error for missing file permissions
  accessSync(path, constants.R_OK | constants.W_OK);
}

export function writeConfig(): boolean {
  // prevent unnecessary rewrites
  if (!hasChanged) return false;
  const stringData = (
    Object.entries(hostData) as Array<[string, string[]]>
  ).reduce((prev, [curAddr, curHosts]) => {
    return `${prev}${curAddr} ${curHosts.join(" ")}\n`;
  }, "");
  writeFileSync(hostConfPath, stringData);
  hasChanged = false;
  return true;
}

// Read initial data from host config backup
function readConfig() {
  const textContent = readFileSync(`${hostConfPath}.${hostConfBakExt}`, {
    encoding: "utf8",
  }) as string;
  // NOTE: maybe properly detect EOL-Style if it causes problems
  // using Unix for now
  for (const line of textContent.split("\n")) {
    //skip empty lines and comments
    if (!line || line.startsWith("#")) continue;
    const { address, hosts } = line.match(/^(?<address>[^\ ]*) (?<hosts>.*)$/)
      ?.groups ?? { address: "172.0.0.1", hosts: "localhost" };
    if (!(address in hostData)) hostData[address] = hosts.split(" ");
    else hostData[address] = hostData[address].concat(hosts.split(" "));
  }
  initialHostData = structuredClone(hostData);
}

export function init() {
  // Boilerplate init code
  // Check permissions for host and backup file
  // conditionally create backup file
  checkFile(hostConfPath);
  try {
    checkFile(`${hostConfPath}.${hostConfBakExt}`);
  } catch (err) {
    if ((err as Error).message.startsWith("No such file or directory:"))
      copyFileSync(hostConfPath, `${hostConfPath}.${hostConfBakExt}`);
    else throw err;
  }

  // load data
  readConfig();
}

function addHost(hosts: string | string[], address = "127.0.0.1") {
  const hostList = typeof hosts === "string" ? hosts.split(" ") : hosts;
  if (!hostData[address]?.some((addr) => hostList.includes(addr))) {
    hostData[address] = (hostData[address] ?? []).concat(hostList);
    hasChanged = true;
  }
}

function removeHost(hosts: string | string[], address = "127.0.0.1") {
  if (address in hostData) {
    const hostList = typeof hosts === "string" ? hosts.split(" ") : hosts;
    hostData[address] = hostData[address].filter(
      (host) => !hostList.includes(host)
    );
    if (hostData[address].length === 0) delete hostData[address];
    hasChanged = true;
  }
}

export function update(dockerHosts: ContainerData[], networkID: string) {
  // Create copy of host list to diff for disconnected containers at the end
  const removedHosts = structuredClone(hostData);

  // Keep all original hosts in list
  for (const address of Object.keys(initialHostData)) {
    removedHosts[address] = removedHosts[address].filter(
      (hostName) => !initialHostData[address].includes(hostName)
    );
  }

  // Add new containers
  for (const container of dockerHosts) {
    if (container.State !== "running" || !(networkID in container.Networks))
      continue;
    const indexAt = removedHosts["127.0.0.1"]?.findIndex(
      (host) => host === container.Name
    );
    if (indexAt !== -1) {
      removedHosts["127.0.0.1"].splice(indexAt, 1);
      continue;
    }
    addHost(container.Name);
  }

  // Remove Containers
  for (const address of Object.keys(removedHosts)) {
    if (removedHosts[address]?.length > 0)
      removeHost(removedHosts[address], address);
  }
}
